/* @(#)devHpe1366a.c	1.112/23/93       */
/* devHpe1366a.c -  */
/*
 * Author:      Frank Lenkszus
 * Date:        03/18/93
 *
 *      Experimental Physics and Industrial Control System (EPICS)
*/
/*
*****************************************************************
                          COPYRIGHT NOTIFICATION
*****************************************************************

THE FOLLOWING IS A NOTICE OF COPYRIGHT, AVAILABILITY OF THE CODE,
AND DISCLAIMER WHICH MUST BE INCLUDED IN THE PROLOGUE OF THE CODE
AND IN ALL SOURCE LISTINGS OF THE CODE.
 
(C)  COPYRIGHT 1993 UNIVERSITY OF CHICAGO
 
Argonne National Laboratory (ANL), with facilities in the States of 
Illinois and Idaho, is owned by the United States Government, and
operated by the University of Chicago under provision of a contract
with the Department of Energy.

Portions of this material resulted from work developed under a U.S.
Government contract and are subject to the following license:  For
a period of five years from March 30, 1993, the Government is
granted for itself and others acting on its behalf a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, and perform
publicly and display publicly.  With the approval of DOE, this
period may be renewed for two additional five year periods. 
Following the expiration of this period or periods, the Government
is granted for itself and others acting on its behalf, a paid-up,
nonexclusive, irrevocable worldwide license in this computer
software to reproduce, prepare derivative works, distribute copies
to the public, perform publicly and display publicly, and to permit
others to do so.

*****************************************************************
                                DISCLAIMER
*****************************************************************

NEITHER THE UNITED STATES GOVERNMENT NOR ANY AGENCY THEREOF, NOR
THE UNIVERSITY OF CHICAGO, NOR ANY OF THEIR EMPLOYEES OR OFFICERS,
MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR ASSUMES ANY LEGAL
LIABILITY OR RESPONSIBILITY FOR THE ACCURACY, COMPLETENESS, OR
USEFULNESS OF ANY INFORMATION, APPARATUS, PRODUCT, OR PROCESS
DISCLOSED, OR REPRESENTS THAT ITS USE WOULD NOT INFRINGE PRIVATELY
OWNED RIGHTS.  

*****************************************************************
LICENSING INQUIRIES MAY BE DIRECTED TO THE INDUSTRIAL TECHNOLOGY
DEVELOPMENT CENTER AT ARGONNE NATIONAL LABORATORY (708-252-2000).
*/
/*
* Modification Log:
 * -----------------
 * .01  03-18-93        frl     initial
 * .02  12-07-93        frl     Ported to R3.11 and ansiified
 * .03	12-14-93	    frl     Changed base Address and upped num cards to 8 
 * .04  05-11-06        ses     Ported to R3.14
 */


#include	<epicsStdlib.h>
#include	<epicsStdioRedirect.h>
#include	<epicsInterrupt.h>
#include	<epicsExport.h>
#include	<string.h>
#include	<sysLib.h>
#include	<vme.h>
#include	<iv.h>
#include	<devLib.h>

#include	<alarm.h>
#include	<dbDefs.h>
#include	<dbAccess.h>
#include    <recSup.h>
#include    <recGbl.h>
#include	<devSup.h>
#include	<dbScan.h>
#include	<module_types.h>
#include	<boRecord.h>
#include	<biRecord.h>
#include	<mbboRecord.h>
#include	<mbbiRecord.h>
#include	<link.h>
#include	<epicsExport.h>

#define         STATIC 		static
STATIC long init();
STATIC long report();
STATIC long init_bo_record(), init_bi_record(), init_mbbi_record(), init_mbbo_record();
STATIC long write_bo(), read_bi(), read_mbbi(), write_mbbo();
STATIC long bi_ioint_info(), mbbi_ioint_info();
STATIC   hpe1366a_getioscanpvt( unsigned short	card, IOSCANPVT	*ppvt);
STATIC long read_card( short card, unsigned long mask, unsigned short  *value,
int reg);
STATIC long write_card( short card, unsigned long mask, unsigned long value,
int reg);
STATIC int checkLink(short card, short signal);
static long initParam(char *parm, unsigned short *pval, unsigned short mode);

int  devHpe1366aDebug = 0;
epicsExportAddress(int,devHpe1366aDebug);

#define	MUX0		     0
#define MUX1		     1
#define NO_ERR_RPT  	-1
#define PARAM_BIT_FIELD  1
#define PARAM_MASK       2


struct hpe1366a {
  unsigned short dev_id;		/* device ID 0xffff */
  unsigned short dev_type;		/* device type 0xff80 or 0xff84 */
  unsigned short csr;			/* csr */
  unsigned short pad1;
  unsigned short mux0;			/* mux 0 control */
  unsigned short mux1;			/* mux 1 control */
  unsigned short pad[26];
};

struct ioCard {
  struct hpe1366a	*card;
  int               lock;
  int			int_vector;
  unsigned short	pending;	/* switch position pending int */
  unsigned short	pending1;	/* switch position pending int */
  unsigned short	shadow;	/* shadow of actual switch position */
  unsigned short	shadow1;	/* shadow of actual switch position */
  short			busy;		/* relays busy */
  IOSCANPVT		*pioscanpvt;
};

#define		MAX_MUX_CARDS	8


#define		HP_ID		0xffff		/* hp id code */
#define		E1366A		0xff80		/* E1366a */
#define		E1367A		0xff84		/* E1367a */

#define         LADD		244             /* logical address */
#define		VXIBASE		0xc000
#define		HPE1366A_BASE  (LADD * 0x40) +  VXIBASE
#define		HPE1366A_INT_LEVEL	1
/* csr command bits */
#define		CSR_W_RESET	0x01		/* software reset */

/* csr status bits */
#define		CSR_R_READY	0x08		/* Module not busy  */

#define		ALL_SWITCHES_OPEN	0

static struct ioCard cards[MAX_MUX_CARDS];
static int		init_flag = 0;

/* Create the dset devBoHpe1366a */
struct {
	long		number;
	DEVSUPFUN	report;		/* used by dbior */
	DEVSUPFUN	init;		/* called 1 time before & after all records */
	DEVSUPFUN	init_record;	/* called 1 time for each record */
	DEVSUPFUN	get_ioint_info;	/* used for COS processing */
	DEVSUPFUN	write_bo;	/* output command goes here */
}devBoHpe1366a={
	5,
	NULL,
	init,
	init_bo_record,
	NULL,
	write_bo
};
epicsExportAddress(dset,devBoHpe1366a);

/* Create the dset devBiHpe1366a */

struct {
        long            number;
        DEVSUPFUN       report;         /* used by dbior */
        DEVSUPFUN       init;           /* called 1 time before & after all reco
rds */
        DEVSUPFUN       init_record;    /* called 1 time for each record */
        DEVSUPFUN       get_ioint_info; /* used for COS processing */
        DEVSUPFUN       read_bi;        /* input command goes here */
}devBiHpe1366a={
        5,
        NULL,
        init,
        init_bi_record,
        bi_ioint_info,
        read_bi
};
epicsExportAddress(dset,devBiHpe1366a);

/* Create the dset devMbboHpe1366a */
struct {
	long		number;
	DEVSUPFUN	report;		/* used by dbior */
	DEVSUPFUN	init;		/* called 1 time before & after all records */
	DEVSUPFUN	init_record;	/* called 1 time for each record */
	DEVSUPFUN	get_ioint_info;	/* used for COS processing */
	DEVSUPFUN	write_mbbo;	/* output command goes here */
}devMbboHpe1366a={
	5,
	NULL,
	init,
	init_mbbo_record,
	NULL,
	write_mbbo
};
epicsExportAddress(dset,devMbboHpe1366a);

/* Create the dset devMbbiHpe1366a */

struct {
        long            number;
        DEVSUPFUN       report;         /* used by dbior */
        DEVSUPFUN       init;           /* called 1 time before & after all reco
rds */
        DEVSUPFUN       init_record;    /* called 1 time for each record */
        DEVSUPFUN       get_ioint_info; /* used for COS processing */
        DEVSUPFUN       read_mbbi;        /* input command goes here */
}devMbbiHpe1366a={
        5,
        NULL,
        init,
        init_mbbi_record,
        mbbi_ioint_info,
        read_mbbi
};
epicsExportAddress(dset,devMbbiHpe1366a);



/**************************************************************************
 *
 *  Interrupt service routine 
 *
***************************************************************************/
void hpe1366a_int( struct ioCard *p)
{
/*
	errMessage(NO_ERR_RPT, "tic\n");
*/
    if ( p == NULL)
		return;
    p->shadow = p->pending;
    p->shadow1 = p->pending1;
    p->busy = FALSE;
    scanIoRequest(*(p->pioscanpvt));
    return;
}


/**************************************************************************
*
* Initialization of HPE1366a RF Mux
*
***************************************************************************/
STATIC long
init( int flag)
{
  char			*xname="devHpe1366a-init";
  int			j;
  unsigned short	card_id;
  unsigned short	model;
  struct hpe1366a	*p;
  char			msg[128];

  if (init_flag != 0)
    return(0);

  init_flag = 1;

  if (sysBusToLocalAdrs(VME_AM_SUP_SHORT_IO, (char *)HPE1366A_BASE,
		 (char **)&p) == -1) {
    if (devHpe1366aDebug)
       printf("devHpe1366a: can not find short address space\n");
    return(-1);
  }

  /* We end up here 1 time before all records are initialized */
  for (j=0; j < MAX_MUX_CARDS; j++) {
    if (devHpe1366aDebug)
  	  printf("devHpe1366a: init card %d\n", j);


    cards[j].card = NULL;		
    if (vxMemProbe((char *)&(p->dev_id) , READ, sizeof(card_id),
		(char *)&card_id) < 0) {
      if (devHpe1366aDebug)
         printf("devHpe1366a: Probe at %p failed\n", &(p->dev_id));
    } else {
      if (devHpe1366aDebug)
         printf("devHpe1366a: Probe at %p success, board id = 0x%2.2X\n", &(p->dev_id), card_id);
      if ( card_id != HP_ID ) {
	sprintf(msg, "%s:  Card installed at %p not an HP \n", xname,
		&(p->dev_id));
	errMessage(NO_ERR_RPT, msg);
      } else  {
    	if (vxMemProbe((char *)&(p->dev_type) , READ, sizeof(model), 
		(char *)&model) < 0) {
    		  if (devHpe1366aDebug) {
         		printf("devHpe1366a: Probe at %p failed\n",
				 &(p->dev_type));
		   }
	} else {
    	  	if (devHpe1366aDebug) {
       		  printf("devHpe1366a: Initializing card at %p\n",
					 &(p->dev_id));
		}
      	    	cards[j].card = p;	/* Remember address of the board */
			/*
			 can't read current state of mux -- set to known 
				state */
		cards[j].int_vector = LADD + j;
			cards[j].pending = ALL_SWITCHES_OPEN; 
			cards[j].pending1 = ALL_SWITCHES_OPEN; 
			cards[j].shadow = ALL_SWITCHES_OPEN;
			cards[j].shadow1 = ALL_SWITCHES_OPEN;
			cards[j].card->mux0 = ALL_SWITCHES_OPEN;
			cards[j].card->mux1 = ALL_SWITCHES_OPEN;
			if ( (cards[j].pioscanpvt = 
			     (IOSCANPVT *)malloc(sizeof(IOSCANPVT))) == NULL) {
    	  			if (devHpe1366aDebug) {
      	printf("devHpe1366a: (init) pioscanpvt malloc  failed for card %d\n", j);
				}
				return(-1);
			}
		        if ( intConnect( INUM_TO_IVEC(cards[j].int_vector),
					 hpe1366a_int, &cards[j]) != 0) {
    	  			if (devHpe1366aDebug) {
       	printf("devHpe1366a: (init) int_Connect failed for card %d\n", j);
				}
				return(-1);
			}
			sysIntEnable(HPE1366A_INT_LEVEL);
			scanIoInit(cards[j].pioscanpvt);
			cards[j].lock = epicsInterruptLock();
   	  		epicsInterruptUnlock(cards[j].lock);  /* Init the board lock */
	}
      }
    }
    p++;
  }
  return(0);
}


/*************************************************************************
 *
 *	bi_ioint_info()
 *
 ************************************************************************/
STATIC  long bi_ioint_info( int	cmd, struct biRecord *pbi, IOSCANPVT *ppvt)
{
	hpe1366a_getioscanpvt(pbi->inp.value.vmeio.card, ppvt);
	return(0);
}
/*************************************************************************
 *
 *	mbbi_ioint_info()
 *
 ************************************************************************/
STATIC  long mbbi_ioint_info(int cmd, struct mbbiRecord *pmbbi, IOSCANPVT *ppvt)
{
	hpe1366a_getioscanpvt(pmbbi->inp.value.vmeio.card, ppvt);
	return(0);
}
/*************************************************************************
 *
 *	hpe1366a_getioscanpvt()
 *
 ************************************************************************/
STATIC   int hpe1366a_getioscanpvt( unsigned short	card, IOSCANPVT	*ppvt)
{
        if ( ( card < MAX_MUX_CARDS) &&
			 ( cards[card].pioscanpvt != NULL)) 
		*ppvt = *(cards[card].pioscanpvt);
	else if (devHpe1366aDebug) {
       	printf("hpe1366a_getioscanpvt: failed for card %d\n", card);
        }
	return(0);
}

/**************************************************************************
 *
 * BO Initialization (Called one time for each BO HPE1366E card record)
 *
 **************************************************************************/
STATIC long init_bo_record( struct boRecord *pbo)
{
    int status=0;
    unsigned short	value, val;
 
    /* bo.out must be an VME_IO */
    switch (pbo->out.type) {
    case (VME_IO) :
/* check param */
      if ( initParam(pbo->out.value.vmeio.parm, &val, PARAM_BIT_FIELD) != 0 ) {
        if (devHpe1366aDebug) {
          errPrintf(NO_ERR_RPT, __FILE__, __LINE__,
                "devHpe1366aBoInit : card = %d Sig = %d  PARM = %s (BAD PARAM)\n",
                   pbo->out.value.vmeio.card, pbo->out.value.vmeio.signal,
			  pbo->out.value.vmeio.parm);
        }
        recGblRecordError(S_dev_badSignal,(void *)pbo,
                "devHpe1366aBoInit (initBoRecord) : init failed!!!");
        status =S_dev_badSignal;
      } else {
         pbo->mask =(unsigned long)val;
         if (read_card(pbo->out.value.vmeio.card, pbo->mask,
				 &value, pbo->out.value.vmeio.signal) == 0)
          		pbo->rbv = pbo->rval = value;
         else status = 2;
      }
      break;
    default :
	pbo->pact = 1;		/* make sure we don't process this thing */
        status = S_db_badField;
        recGblRecordError(status,(void *)pbo,
            "devHpe1366a (init_record) Illegal OUT field");
    }
    if (status == 0)
	pbo->dpvt = pbo;
    return(status);
}


/**************************************************************************
 *
 * BI Initialization (Called one time for each BI HPE1366E card record)
 *
 **************************************************************************/
STATIC long init_bi_record( struct biRecord *pbi)
{
  int status=0;
  unsigned short	value, val;

  switch (pbi->inp.type) {
  case (VME_IO) :
/* check param */
    if ( initParam(pbi->inp.value.vmeio.parm, &val, PARAM_BIT_FIELD) != 0 ) {
      if (devHpe1366aDebug) {
        errPrintf(NO_ERR_RPT, __FILE__, __LINE__,
                "devHpe1366aBiInit : card = %d Sig = %d  PARM = %s (BAD PARAM)\n",
                   pbi->inp.value.vmeio.card, pbi->inp.value.vmeio.signal,
			  pbi->inp.value.vmeio.parm);
        }
        recGblRecordError(S_dev_badSignal,(void *)pbi,
                "devHpe1366aBiInit (initBiRecord) : init failed!!!");
        status =S_dev_badSignal;
    } else {
       pbi->mask =(unsigned long)val;
       if (read_card(pbi->inp.value.vmeio.card, pbi->mask,
				 &value, pbi->inp.value.vmeio.signal) == 0)
        		pbi->rval = value;
       else status = 2;
    }
    break;
  default:
    pbi->pact = 1;		/* make sure we don't process this thing */
    status = S_db_badField;
    recGblRecordError(status,(void *)pbi,
        "devHpe1366a (init_record) Illegal INP field");
  }

    if (status == 0)
	pbi->dpvt = pbi;
  return(status);
}


/**************************************************************************
 *
 * Perform a write operation from a BO record
 *
 **************************************************************************/
STATIC long write_bo( struct boRecord *pbo)
{


  if (!pbo->dpvt)
	return(S_dev_NoInit);
  if (write_card(pbo->out.value.vmeio.card, pbo->mask, pbo->rval,
		 pbo->out.value.vmeio.signal) == 0) {
	pbo->rbv = pbo->rval;	/* write only register so fake it */
  	return(0);
  }
  /* Set an alarm for the record */
  recGblSetSevr(pbo, WRITE_ALARM, INVALID_ALARM);
  return(2);
}

/**************************************************************************
 *
 * Perform a read operation from a BI record
 *
 **************************************************************************/
STATIC long read_bi( struct biRecord *pbi)
{

  unsigned short	value;

  if (!pbi->dpvt)
	return(S_dev_NoInit);

   if (read_card(pbi->inp.value.vmeio.card, pbi->mask, &value,
		 pbi->inp.value.vmeio.signal) == 0) {
   	pbi->rval = value;
    	return(0);
    }
  /* Set an alarm for the record */
  recGblSetSevr(pbi, READ_ALARM, INVALID_ALARM);
  return(2);
}


/**************************************************************************
 *
 * MBBO Initialization (Called one time for each MBBO HPE1366E card record)
 *
 **************************************************************************/
STATIC long init_mbbo_record( struct mbboRecord *pbo)
{
    int status=0;
    unsigned short	value;
 
    /* mbbo.out must be an VME_IO */
    switch (pbo->out.type) {
    case (VME_IO) :
        pbo->shft = 0; 
        pbo->mask <<= pbo->shft;
       	if (read_card(pbo->out.value.vmeio.card, pbo->mask,
  		 	&value, pbo->out.value.vmeio.signal) == 0) 
       		pbo->rbv = pbo->rval = value;
       	else {
	if (devHpe1366aDebug >= 10) 
   	   printf("devHpe1366a: (init_mbbo_record) Init on card  %d failed\n",
			 pbo->out.value.vmeio.card);
		status = 2;
	}
        break;
    default :
	pbo->pact = 1;		/* make sure we don't process this thing */
        status = S_db_badField;
        recGblRecordError(status,(void *)pbo,
            "devMBBOHpe1366a (init_record) Illegal OUT field");
    }
    if (status == 0)
	pbo->dpvt = pbo;
    return(status);
}


/**************************************************************************
 *
 * MBBI Initialization (Called one time for each MBBI HPE1366E card record)
 *
 **************************************************************************/
STATIC long init_mbbi_record( struct mbbiRecord *pbi)
{
  int status=0;
  unsigned short	value;

  switch (pbi->inp.type) {
  case (VME_IO) :
/*
    pbi->shft = (unsigned short) pbi->inp.value.vmeio.signal; 
*/
    pbi->shft = 0; 
    pbi->mask <<= (unsigned long)pbi->shft;
   if (read_card(pbi->inp.value.vmeio.card, pbi->mask, &value, 
			pbi->inp.value.vmeio.signal) == 0) {
      		pbi->rval = (unsigned long)value;
   } else {
	if (devHpe1366aDebug >= 10) 
   	   printf("devHpe1366a: (init_mbbi_record) Init on card  %d failed\n",
			 pbi->inp.value.vmeio.card);
		status = 2;
   }
    break;
  default:
    pbi->pact = 1;		/* make sure we don't process this thing */
    status = S_db_badField;
    recGblRecordError(status,(void *)pbi,
        "devMBBIHpe1366a (init_record) Illegal INP field");
  }
    if (status == 0)
	pbi->dpvt = pbi;
  return(status);
}


/**************************************************************************
 *
 * Perform a write operation from a MBBO record
 *
 **************************************************************************/
STATIC long write_mbbo( struct mbboRecord *pbo)
{

  if (!pbo->dpvt)
	return(S_dev_NoInit);
  if (write_card(pbo->out.value.vmeio.card, pbo->mask, pbo->rval,
			 pbo->out.value.vmeio.signal) == 0)
      	return(0);
	
  /* Set an alarm for the record */
  recGblSetSevr(pbo, WRITE_ALARM, INVALID_ALARM);
  return(2);
}

/**************************************************************************
 *
 * Perform a read operation from a MBBI record
 *
 **************************************************************************/
STATIC long read_mbbi( struct mbbiRecord *pbi)
{

  unsigned short	value;
  if (!pbi->dpvt)
	return(S_dev_NoInit);
  if (read_card(pbi->inp.value.vmeio.card, pbi->mask, &value, 
			pbi->inp.value.vmeio.signal) == 0) {
    		pbi->rval = value;
    		return(0);
  }
  /* Set an alarm for the record */
  recGblSetSevr(pbi, READ_ALARM, INVALID_ALARM);
  return(2);
}


/**************************************************************************
 *
 * Raw read a bitfield from the card
 *
 **************************************************************************/
STATIC long read_card(
short           card,   /* Link number from DCT */
unsigned long   mask,   /* created in init_bo_record() */
unsigned short  *value, /* the value to return from the card */
int             reg)
{
  if (checkLink(card, 0) == -1)
    return(-1);

  if (reg == MUX1)
    *value = cards[card].shadow1 & mask;
  else
    *value = cards[card].shadow & mask;

  if (devHpe1366aDebug >= 20)
    printf("devHpe1366a: read 0x%4.4X from card %d\n", *value, card);

  return(0);
}

/**************************************************************************
 *
 * Raw write a bitfield to the card
 *
 **************************************************************************/
STATIC long write_card( short card, unsigned long mask, unsigned long value,
int reg)
{
unsigned	short	work;

  if (checkLink(card, 0) == -1) {
    return(-1);
  }


  cards[card].lock =epicsInterruptLock();
  if (reg == MUX1) {
  	work = cards[card].pending1;
 	work = (work & ~mask) | ( value & mask);
  	cards[card].card->mux1 = work;
  	cards[card].pending1 = work;
  } else {
  	work = cards[card].pending ;
  	work = (work & ~mask) | ( value & mask);
  	cards[card].card->mux0 = work;
  	cards[card].pending = work;
  }
  cards[card].busy = TRUE;
  epicsInterruptUnlock(cards[card].lock);

  if (devHpe1366aDebug >= 10)
    printf("devHpe1366a: wrote 0x%4.4X to card %d\n", cards[card].pending, card);

  return(0);
}


/**************************************************************************
 *
 * Make sure card number is valid
 *
 **************************************************************************/
STATIC int checkLink( short card, short signal)
{
  if (card >= MAX_MUX_CARDS || card < 0) {
    if (devHpe1366aDebug >= 10)
       printf("devHpe1366a: bad card # of %d\n", card);
    return(-1);
  }
  if (cards[card].card == NULL) {
    if (devHpe1366aDebug >= 10)
       printf("devHpe1366a: Address for card %d is NULL\n", card);
    return(-1);
  }

  return(0);
}

static long initParam(char *parm, unsigned short *pval, unsigned short mode)
{
	unsigned short val;
	unsigned short	val1;
	char 	*xname="initParam";

  if( parm == NULL) {
      if (devHpe1366aDebug) {
     	errPrintf(NO_ERR_RPT, __FILE__, __LINE__,
		"%s: NULL pointer for Parameter String", xname);
	}
      return(-1);
  }
  val = 1;
  if( mode == PARAM_BIT_FIELD ) {
	if ( *parm >= '0' && *parm <= '9') {
		*pval = val << (*parm - '0');
		return(0);
	} else if ( *parm >= 'a' && *parm <= 'f') {
		*pval = val << ( *parm - 'a' + 10 );
		return(0);
	} else if ( *parm >= 'A' && *parm <= 'F') {
		*pval = val << ( *parm - 'A' + 10 );
		return(0);
	} else
		return(-1);
  } else if ( mode == PARAM_MASK) {
	if ( sscanf(parm, "%hx", &val1) != 1 ) {
      		if (devHpe1366aDebug) {
     			errPrintf(NO_ERR_RPT, __FILE__, __LINE__,
				"%s: Bad Parameter = %s ", xname, parm);
		}
      		return(-1);
        }
	*pval = (unsigned short)val1;
	return(0);
  } else
	return(-1);

}
